from typing import Any, Dict, Optional, Type

from rest_framework.serializers import Serializer
from rest_framework.views import APIView


class BaseAPIView(APIView):
    serializer_class: Optional[Type[Serializer]] = None

    def get_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        deserializing input.
        """
        serializer_class = self.get_serializer_class()
        kwargs.setdefault('context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)  # pylint: disable=not-callable

    def get_serializer_class(self) -> Type[Serializer]:
        """
        Return the class to use for the serializer.
        Defaults to using `self.serializer_class`.

        You may want to override this if you need to provide different
        serializations depending on the incoming request.

        (Eg. admins get full serialization, others get basic serialization)
        """
        assert self.serializer_class is not None, (
            "'%s' should either include a `serializer_class` attribute, "
            "or override the `get_serializer_class()` method."
            % self.__class__.__name__
        )

        return self.serializer_class

    def get_serializer_context(self) -> Dict[str, Any]:
        """
        Extra context provided to the serializer class.
        """
        return {
            'request': self.request,
            'view': self
        }



from typing import TYPE_CHECKING, Any, Dict, Type

from django.contrib import auth
from django.utils.translation import gettext as _
from rest_framework import permissions, serializers
from rest_framework.authentication import SessionAuthentication, TokenAuthentication
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework.settings import api_settings

from rest_registration.api.views.base import BaseAPIView
from rest_registration.exceptions import LoginInvalid, UserNotFound
from rest_registration.settings import registration_settings
from rest_registration.utils.auth_backends import get_login_authentication_backend
from rest_registration.utils.responses import get_ok_response

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser


class LoginView(BaseAPIView):
    permission_classes = registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES

    def post(self, request: Request) -> Response:
        '''
        Logs in the user via given login and password.
        '''
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        login_authenticator = registration_settings.LOGIN_AUTHENTICATOR
        try:
            user = login_authenticator(serializer.validated_data, serializer=serializer)
        except UserNotFound:
            raise LoginInvalid() from None

        extra_data = perform_login(request, user)

        return get_ok_response(_("Login successful"), extra_data=extra_data)

    def get_serializer_class(self) -> Type[Serializer]:
        return registration_settings.LOGIN_SERIALIZER_CLASS


login = LoginView.as_view()


class LogoutSerializer(serializers.Serializer):  # pylint: disable=abstract-method
    revoke_token = serializers.BooleanField(default=False)


class LogoutView(BaseAPIView):
    serializer_class = LogoutSerializer
    permission_classes = [permissions.IsAuthenticated]

    def post(self, request: Request) -> Response:
        '''
        Logs out the user. returns an error if the user is not
        authenticated.
        '''
        user = request.user
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = serializer.validated_data

        if should_authenticate_session():
            auth.logout(request)
        if should_retrieve_token() and data['revoke_token']:
            auth_token_manager_cls = registration_settings.AUTH_TOKEN_MANAGER_CLASS
            auth_token_manager = auth_token_manager_cls()  # noqa: E501 type: rest_registration.auth_token_managers.AbstractAuthTokenManager
            auth_token_manager.revoke_token(user)

        return get_ok_response(_("Logout successful"))


logout = LogoutView.as_view()


def should_authenticate_session() -> bool:
    result = registration_settings.LOGIN_AUTHENTICATE_SESSION
    if result is None:
        result = rest_auth_has_class(SessionAuthentication)
    return result


def should_retrieve_token() -> bool:
    result = registration_settings.LOGIN_RETRIEVE_TOKEN
    if result is None:
        result = rest_auth_has_class(TokenAuthentication)
    return result


def rest_auth_has_class(cls: type) -> bool:
    return cls in api_settings.DEFAULT_AUTHENTICATION_CLASSES


def perform_login(request: Request, user: 'AbstractBaseUser') -> Dict[str, Any]:
    if should_authenticate_session():
        login_auth_backend = get_login_authentication_backend(user=user)
        auth.login(request, user, backend=login_auth_backend)

    extra_data = {}

    if should_retrieve_token():
        auth_token_manager_cls = registration_settings.AUTH_TOKEN_MANAGER_CLASS
        auth_token_manager = auth_token_manager_cls()  # noqa: E501 type: rest_registration.auth_token_managers.AbstractAuthTokenManager
        token = auth_token_manager.provide_token(user)
        extra_data['token'] = token

    return extra_data



from django.test import TestCase
from django.test.utils import override_settings

from rest_registration.utils.nested_settings import NestedSettings


class RegistrationSettingsTestCase(TestCase):

    def setUp(self):
        self.defaults = {
            'A': 2,
            'B': 3,
        }

    def test_user_settings(self):
        user_settings = {
            'A': 1,
        }
        settings = NestedSettings(
            user_settings, self.defaults, (), 'NESTED_TEST_SETTING')
        self.assertEqual(settings.A, 1)
        self.assertEqual(settings.B, 3)

    @override_settings(
        REST_REGISTRATION={
            'A': 5,
        }
    )
    def test_django_settings(self):
        settings = NestedSettings(
            None, self.defaults, (), 'REST_REGISTRATION')
        self.assertEqual(settings.A, 5)
        self.assertEqual(settings.B, 3)



from django.urls import path

from .views import reset_password, verify_email, verify_registration

app_name = 'rest_registration.contrib.verification_redirects'
urlpatterns = [
    path('verify-registration/', verify_registration, name='verify-registration'),
    path('verify-email/', verify_email, name='verify-email'),
    path('reset-password/', reset_password, name='reset-password'),
]



from typing import Type

from django.http import Http404
from django.utils.translation import gettext as _
from rest_framework import serializers
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import Serializer

from rest_registration.api.serializers import PasswordConfirmSerializerMixin
from rest_registration.api.views.base import BaseAPIView
from rest_registration.exceptions import UserNotFound
from rest_registration.settings import registration_settings
from rest_registration.signers.reset_password import ResetPasswordSigner
from rest_registration.utils.responses import get_ok_response
from rest_registration.utils.users import get_user_by_verification_id
from rest_registration.utils.validation import (
    run_validators,
    validate_password_with_user_id,
    validate_user_password_confirm
)
from rest_registration.utils.verification import verify_signer_or_bad_request


class SendResetPasswordLinkView(BaseAPIView):
    permission_classes = registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES

    def post(self, request: Request) -> Response:
        '''
        Send email with reset password link.
        '''
        if not registration_settings.RESET_PASSWORD_VERIFICATION_ENABLED:
            raise Http404()
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        if registration_settings.RESET_PASSWORD_FAIL_WHEN_USER_NOT_FOUND:
            success_message = _("Reset link sent")
        else:
            success_message = _("Reset link sent if the user exists in database")
        user_finder = registration_settings.SEND_RESET_PASSWORD_LINK_USER_FINDER
        try:
            user = user_finder(serializer.validated_data, serializer=serializer)
        except UserNotFound:
            if registration_settings.RESET_PASSWORD_FAIL_WHEN_USER_NOT_FOUND:
                raise
            return get_ok_response(success_message)
        email_sender = registration_settings.RESET_PASSWORD_VERIFICATION_EMAIL_SENDER
        email_sender(request, user)
        return get_ok_response(success_message)

    def get_serializer_class(self) -> Type[Serializer]:
        return registration_settings.SEND_RESET_PASSWORD_LINK_SERIALIZER_CLASS


send_reset_password_link = SendResetPasswordLinkView.as_view()


class ResetPasswordSerializer(  # pylint: disable=abstract-method
        PasswordConfirmSerializerMixin,
        serializers.Serializer):
    user_id = serializers.CharField(required=True)
    timestamp = serializers.IntegerField(required=True)
    signature = serializers.CharField(required=True)
    password = serializers.CharField(required=True)

    def has_password_confirm_field(self):
        return registration_settings.RESET_PASSWORD_SERIALIZER_PASSWORD_CONFIRM

    def validate(self, attrs):
        validators = [
            validate_password_with_user_id,
        ]
        if self.has_password_confirm_field():
            validators.append(validate_user_password_confirm)
        run_validators(validators, attrs)
        return attrs


class ResetPassordView(BaseAPIView):
    serializer_class = ResetPasswordSerializer
    permission_classes = registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES

    def post(self, request: Request) -> Response:
        '''
        Reset password, given the signature and timestamp from the link.
        '''
        process_reset_password_data(
            request.data,
            serializer_context=self.get_serializer_context(),
        )
        return get_ok_response(_("Reset password successful"))


reset_password = ResetPassordView.as_view()


def process_reset_password_data(input_data, serializer_context=None):
    if serializer_context is None:
        serializer_context = {}
    if not registration_settings.RESET_PASSWORD_VERIFICATION_ENABLED:
        raise Http404()
    serializer = ResetPasswordSerializer(data=input_data, context=serializer_context)
    serializer.is_valid(raise_exception=True)

    data = serializer.validated_data.copy()
    password = data.pop('password')
    data.pop('password_confirm', None)
    # We use the signer only for verification, therefore we don't need a base_url and
    # may set strict=False
    signer = ResetPasswordSigner(data, strict=False)
    verify_signer_or_bad_request(signer)

    user = get_user_by_verification_id(data['user_id'], require_verified=False)
    user.set_password(password)
    user.save()



from django.apps import apps
from django.core.checks.registry import registry
from django.test import TestCase
from django.test.utils import override_settings

from rest_registration.auth_token_managers import AbstractAuthTokenManager
from rest_registration.checks import ErrorCode, WarningCode
from rest_registration.settings import DEFAULTS
from tests.helpers.settings import override_rest_registration_settings


def simulate_checks():
    app_configs = apps.app_configs
    errors = []
    all_checks = registry.get_checks(False)
    rest_registration_checks = [
        check for check in all_checks
        if check.__module__.startswith('rest_registration.')
    ]
    for check in rest_registration_checks:
        errors.extend(check(app_configs))
    return errors


class ChecksTestCase(TestCase):

    @override_settings(
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_URL': '/verify-account/',
            'REGISTER_EMAIL_VERIFICATION_URL': '/verify-email/',
            'RESET_PASSWORD_VERIFICATION_URL': '/reset-password/',
        },
    )
    def test_checks_preferred_setup_missing_sender_email(self):
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.NO_VER_FROM_EMAIL,
        ])

    @override_settings(
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_URL': '/verify-account/',
            'REGISTER_VERIFICATION_AUTO_LOGIN': True,
            'REGISTER_EMAIL_VERIFICATION_ENABLED': False,
            'RESET_PASSWORD_VERIFICATION_ENABLED': False,
            'VERIFICATION_FROM_EMAIL': 'jon.doe@example.com',
        },
    )
    def test_checks_multiple_time_auto_login(self):
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            WarningCode.REGISTER_VERIFICATION_MULTIPLE_AUTO_LOGIN,
        ])

    @override_settings(
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_ENABLED': False,
            'REGISTER_EMAIL_VERIFICATION_ENABLED': False,
            'RESET_PASSWORD_VERIFICATION_URL': '/reset-password/',
        },
    )
    def test_checks_one_verification_url_missing_sender_email(self):
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.NO_VER_FROM_EMAIL,
        ])

    @override_settings(
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_URL': '/verify-account/',
            'REGISTER_EMAIL_VERIFICATION_URL': '/verify-email/',
            'RESET_PASSWORD_VERIFICATION_URL': '/reset-password/',
            'VERIFICATION_FROM_EMAIL': 'jon.doe@example.com',
            'LOGIN_RETRIEVE_TOKEN': True,
        },
    )
    def test_checks_missing_token_auth_config(self):
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.NO_TOKEN_AUTH_CONFIG,
        ])

    @override_settings(
        INSTALLED_APPS=(
            'django.contrib.admin',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.messages',
            'django.contrib.staticfiles',

            'rest_framework',
            'rest_registration',
        ),
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_ENABLED': False,
            'REGISTER_EMAIL_VERIFICATION_ENABLED': False,
            'RESET_PASSWORD_VERIFICATION_ENABLED': False,
        },
    )
    def test_checks_missing_auth_installed(self):
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.NO_AUTH_INSTALLED,
        ])

    @override_settings(
        INSTALLED_APPS=(
            'django.contrib.admin',
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.messages',
            'django.contrib.staticfiles',

            'rest_framework',
            'rest_registration',
        ),
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_URL': '/verify-account/',
            'REGISTER_EMAIL_VERIFICATION_URL': '/verify-email/',
            'RESET_PASSWORD_VERIFICATION_URL': '/reset-password/',
            'VERIFICATION_FROM_EMAIL': 'jon.doe@example.com',
            'LOGIN_RETRIEVE_TOKEN': True,
        },
    )
    def test_checks_missing_token_auth_installed(self):
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.NO_TOKEN_AUTH_CONFIG,
            ErrorCode.NO_TOKEN_AUTH_INSTALLED,
        ])

    @override_settings(
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_URL': '/verify-account/',
            'REGISTER_EMAIL_VERIFICATION_URL': '/verify-email/',
            'RESET_PASSWORD_VERIFICATION_URL': '/reset-password/',
            'VERIFICATION_FROM_EMAIL': 'jon.doe@example.com',
            'REGISTER_VERIFICATION_EMAIL_TEMPLATES': {},
        },
    )
    def test_checks_invalid_register_verification_email_templates_config(self):
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.INVALID_EMAIL_TEMPLATE_CONFIG,
        ])

    @override_settings(
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_URL': '/verify-account/',
            'REGISTER_EMAIL_VERIFICATION_URL': '/verify-email/',
            'RESET_PASSWORD_VERIFICATION_URL': '/reset-password/',
            'VERIFICATION_FROM_EMAIL': 'jon.doe@example.com',
            'RESET_PASSWORD_VERIFICATION_EMAIL_TEMPLATES': {},
        },
    )
    def test_checks_invalid_register_email_verification_email_templates_config(self):  # noqa: E501
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.INVALID_EMAIL_TEMPLATE_CONFIG,
        ])

    @override_settings(
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_URL': '/verify-account/',
            'REGISTER_EMAIL_VERIFICATION_URL': '/verify-email/',
            'RESET_PASSWORD_VERIFICATION_URL': '/reset-password/',
            'VERIFICATION_FROM_EMAIL': 'jon.doe@example.com',
            'RESET_PASSWORD_VERIFICATION_EMAIL_TEMPLATES': {},
        },
    )
    def test_checks_invalid_reset_password_verification_email_templates_config(self):  # noqa: E501
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.INVALID_EMAIL_TEMPLATE_CONFIG,
        ])

    @override_settings(
        AUTHENTICATION_BACKENDS=[
            'django.contrib.auth.backends.AllowAllUsersModelBackend',
        ],
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_ENABLED': False,
            'REGISTER_EMAIL_VERIFICATION_ENABLED': False,
            'RESET_PASSWORD_VERIFICATION_ENABLED': False,
        },
    )
    def test_checks_invalid_auth_model_backend_used(self):
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.DRF_INCOMPATIBLE_DJANGO_AUTH_BACKEND,
        ])

    @override_settings(
        AUTHENTICATION_BACKENDS=[
            'django.contrib.auth.backends.AllowAllUsersRemoteUserBackend',
        ],
        REST_REGISTRATION={
            'REGISTER_VERIFICATION_ENABLED': False,
            'REGISTER_EMAIL_VERIFICATION_ENABLED': False,
            'RESET_PASSWORD_VERIFICATION_ENABLED': False,
        },
    )
    def test_checks_invalid_auth_remote_backend_used(self):
        errors = simulate_checks()
        self.assert_error_codes_match(errors, [
            ErrorCode.DRF_INCOMPATIBLE_DJANGO_AUTH_BACKEND,
        ])

    def assert_error_codes_match(self, errors, expected_error_codes):
        assert_error_codes_match(errors, expected_error_codes)


@override_settings(REST_REGISTRATION=DEFAULTS)
def test_fail_when_default():
    errors = simulate_checks()
    assert_error_codes_match(errors, [
        ErrorCode.NO_REGISTER_EMAIL_VER_URL,
        ErrorCode.NO_REGISTER_VER_URL,
        ErrorCode.NO_RESET_PASSWORD_VER_URL,
        ErrorCode.NO_VER_FROM_EMAIL,
    ])


@override_rest_registration_settings({
    'REGISTER_VERIFICATION_ENABLED': False,
    'REGISTER_EMAIL_VERIFICATION_ENABLED': False,
    'RESET_PASSWORD_VERIFICATION_ENABLED': False,
})
def test_ok_when_minimal_setup():
    errors = simulate_checks()
    assert_error_codes_match(errors, [])


@override_settings(
    TEMPLATES=(),
)
@override_rest_registration_settings({
    'REGISTER_VERIFICATION_ENABLED': False,
    'REGISTER_EMAIL_VERIFICATION_ENABLED': False,
    'RESET_PASSWORD_VERIFICATION_ENABLED': False,
})
def test_ok_when_no_templates_minimal_setup():
    errors = simulate_checks()
    assert_error_codes_match(errors, [])


def test_ok_when_preferred_setup(
        settings_with_register_verification,
        settings_with_register_email_verification,
        settings_with_reset_password_verification,
):
    errors = simulate_checks()
    assert_error_codes_match(errors, [])


@override_settings(
    TEMPLATES=(),
)
def test_fail_when_no_templates_preferred_setup(
        settings_with_register_verification,
        settings_with_register_email_verification,
        settings_with_reset_password_verification,
):
    errors = simulate_checks()
    assert_error_codes_match(errors, [
        ErrorCode.INVALID_EMAIL_TEMPLATE_CONFIG,
        ErrorCode.INVALID_EMAIL_TEMPLATE_CONFIG,
        ErrorCode.INVALID_EMAIL_TEMPLATE_CONFIG,
    ])

    expected_messages = {
        "REGISTER_VERIFICATION_EMAIL_TEMPLATES is invalid: Template 'rest_registration/register/subject.txt' does not exist; ensure that your Django TEMPLATES setting is configured correctly",  # noqa: E501
        "REGISTER_EMAIL_VERIFICATION_EMAIL_TEMPLATES is invalid: Template 'rest_registration/register_email/subject.txt' does not exist; ensure that your Django TEMPLATES setting is configured correctly",  # noqa: E501
        "RESET_PASSWORD_VERIFICATION_EMAIL_TEMPLATES is invalid: Template 'rest_registration/reset_password/subject.txt' does not exist; ensure that your Django TEMPLATES setting is configured correctly",  # noqa: E501
    }
    assert {e.msg for e in errors} == expected_messages


@override_rest_registration_settings({
    'USER_LOGIN_FIELDS': ['username'],
})
def test_when_one_unique_login_field_then_check_succeeds():
    errors = simulate_checks()
    assert_error_codes_match(errors, [])


@override_rest_registration_settings({
    'USER_LOGIN_FIELDS': ['email'],
})
def test_when_one_non_unique_login_field_then_check_fails():
    errors = simulate_checks()
    assert_error_codes_match(errors, [
        ErrorCode.LOGIN_FIELDS_NOT_UNIQUE,
    ])


@override_rest_registration_settings({
    'USER_LOGIN_FIELDS': ['username', 'email'],
})
def test_when_one_non_unique_login_field_in_many_then_check_fails():
    errors = simulate_checks()
    assert_error_codes_match(errors, [
        ErrorCode.LOGIN_FIELDS_NOT_UNIQUE,
    ])


@override_rest_registration_settings({
    'AUTH_TOKEN_MANAGER_CLASS': 'tests.testapps.custom_authtokens.auth.FaultyAuthTokenManager',  # noqa: E501
})
def test_when_custom_authtokenmanager_wrt_specs_then_check_succeeds():
    errors = simulate_checks()
    assert_error_codes_match(errors, [])


class InvalidAuthTokenManager:
    pass


class NotImplementedAuthTokenManager(AbstractAuthTokenManager):  # noqa: E501 pylint: disable=abstract-method
    pass


@override_rest_registration_settings({
    'AUTH_TOKEN_MANAGER_CLASS': InvalidAuthTokenManager,
})
def test_when_authtokenmanager_is_not_correct_subclass_then_check_fails():
    errors = simulate_checks()
    assert_error_codes_match(errors, [
        ErrorCode.INVALID_AUTH_TOKEN_MANAGER_CLASS,
    ])


@override_rest_registration_settings({
    'AUTH_TOKEN_MANAGER_CLASS': NotImplementedAuthTokenManager,
})
def test_when_authtokenmanager_does_not_implement_methods_then_check_fails():
    errors = simulate_checks()
    assert_error_codes_match(errors, [
        ErrorCode.INVALID_AUTH_TOKEN_MANAGER_CLASS,
        ErrorCode.INVALID_AUTH_TOKEN_MANAGER_CLASS,
    ])
    expected_messages = {
        "AUTH_TOKEN_MANAGER_CLASS is not implementing method get_authentication_class",  # noqa: E501
        "AUTH_TOKEN_MANAGER_CLASS is not implementing method provide_token",
    }
    assert {e.msg for e in errors} == expected_messages


@override_settings(
    AUTHENTICATION_BACKENDS=[
        'django.contrib.auth.backends.ModelBackend',
        'django.contrib.auth.backends.RemoteUserBackend',
    ],
)
def test_ok_when_multiple_auth_backends():
    errors = simulate_checks()
    assert_error_codes_match(errors, [])


@override_settings(
    AUTHENTICATION_BACKENDS=[
        'tests.testapp.custom_auth_backends.backend.CustomAuthBackend',
        'django.contrib.auth.backends.RemoteUserBackend',
    ],
)
def test_ok_when_multiple_custom_auth_backends():
    errors = simulate_checks()
    assert_error_codes_match(errors, [])


@override_rest_registration_settings({
    'LOGIN_DEFAULT_SESSION_AUTHENTICATION_BACKEND': 'django.contrib.auth.backends.RemoteUserBackend',  # noqa: E501
})
@override_settings(
    AUTHENTICATION_BACKENDS=[
        'django.contrib.auth.backends.ModelBackend',
        'django.contrib.auth.backends.RemoteUserBackend',
    ],
)
def test_ok_when_login_auth_backend_in_multiple_auth_backends():
    errors = simulate_checks()
    assert_error_codes_match(errors, [])


@override_rest_registration_settings({
    'LOGIN_DEFAULT_SESSION_AUTHENTICATION_BACKEND': 'nonexistent.backend',
})
@override_settings(
    AUTHENTICATION_BACKENDS=[
        'django.contrib.auth.backends.ModelBackend',
        'django.contrib.auth.backends.RemoteUserBackend',
    ],
)
def test_fail_when_login_auth_backend_not_in_multiple_auth_backends():
    errors = simulate_checks()
    assert_error_codes_match(errors, [
        ErrorCode.INVALID_AUTH_BACKENDS_CONFIG,
    ])
    expected_messages = {
        "invalid authentication backends configuration:"
        " LOGIN_DEFAULT_SESSION_AUTHENTICATION_BACKEND"
        " is not in AUTHENTICATION_BACKENDS",
    }
    assert {e.msg for e in errors} == expected_messages


@override_rest_registration_settings({
    'REGISTER_EMAIL_SERIALIZER_CLASS': 'tests.testapps.custom_serializers.serializers.InvalidRegisterEmailSerializer',  # noqa: E501
})
def test_invalid_register_email_serializer_check_fails():
    errors = simulate_checks()
    assert_error_codes_match(errors, [
        ErrorCode.INVALID_REGISTER_EMAIL_SERIALIZER_CLASS,
    ])
    expected_messages = {
        "REGISTER_EMAIL_SERIALIZER_CLASS does not contain email field",
    }
    assert {e.msg for e in errors} == expected_messages


@override_rest_registration_settings({
    'SEND_RESET_PASSWORD_LINK_SERIALIZER_USE_EMAIL': True,
})
def test_fail_when_send_reset_password_link_serializer_uses_non_unique_email():
    errors = simulate_checks()
    assert_error_codes_match(errors, [
        ErrorCode.NON_UNIQUE_FIELD_USED_AS_UNIQUE,
    ])
    expected_messages = {
        "SEND_RESET_PASSWORD_LINK_SERIALIZER_USE_EMAIL is set but email field is not unique",  # noqa: E501
    }
    assert {e.msg for e in errors} == expected_messages


def assert_error_codes_match(errors, expected_error_codes):
    error_ids = sorted(e.id for e in errors)
    expected_error_ids = sorted(
        code.get_full_code_id() for code in expected_error_codes)
    msg = "\n\nList of errors:\n"
    for error in errors:
        msg += "- {error}\n".format(error=error)
    msg += " does not match the codes: "
    if expected_error_codes:
        msg += ", ".join(str(e) for e in expected_error_codes)
    else:
        msg += "(empty list)"
    assert error_ids == expected_error_ids, msg



from django.contrib.auth.password_validation import validate_password
from django.utils.translation import gettext as _
from rest_framework import permissions, serializers
from rest_framework.request import Request
from rest_framework.response import Response

from rest_registration.api.serializers import PasswordConfirmSerializerMixin
from rest_registration.api.views.base import BaseAPIView
from rest_registration.settings import registration_settings
from rest_registration.utils.responses import get_ok_response
from rest_registration.utils.validation import validate_user_password_confirm


class ChangePasswordSerializer(  # pylint: disable=abstract-method
        PasswordConfirmSerializerMixin,
        serializers.Serializer):
    old_password = serializers.CharField()
    password = serializers.CharField()

    def has_password_confirm_field(self):
        return registration_settings.CHANGE_PASSWORD_SERIALIZER_PASSWORD_CONFIRM

    def validate_old_password(self, old_password):
        user = self.context['request'].user
        if not user.check_password(old_password):
            raise serializers.ValidationError(_("Old password is not correct"))
        return old_password

    def validate_password(self, password):
        user = self.context['request'].user
        validate_password(password, user=user)
        return password

    def validate(self, attrs):
        if self.has_password_confirm_field():
            validate_user_password_confirm(attrs)
        return attrs


class ChangePasswordView(BaseAPIView):
    serializer_class = ChangePasswordSerializer
    permission_classes = [permissions.IsAuthenticated]

    def post(self, request: Request) -> Response:
        '''
        Change the user password.
        '''
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        user = request.user
        user.set_password(serializer.validated_data['password'])
        user.save()
        return get_ok_response(_("Password changed successfully"))


change_password = ChangePasswordView.as_view()



from typing import Type

from django.db import transaction
from django.http import Http404
from django.utils.translation import gettext as _
from rest_framework import serializers, status
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import Serializer

from rest_registration import signals
from rest_registration.api.views.base import BaseAPIView
from rest_registration.api.views.login import perform_login
from rest_registration.exceptions import UserWithoutEmailNonverifiable
from rest_registration.settings import registration_settings
from rest_registration.signers.register import RegisterSigner
from rest_registration.utils.responses import get_ok_response
from rest_registration.utils.users import (
    get_user_by_verification_id,
    get_user_email_field_name,
    get_user_setting
)
from rest_registration.utils.verification import verify_signer_or_bad_request


class RegisterView(BaseAPIView):
    permission_classes = registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES

    def post(self, request: Request) -> Response:
        '''
        Register new user.
        '''
        if not registration_settings.REGISTER_FLOW_ENABLED:
            raise Http404()
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        kwargs = {}

        if registration_settings.REGISTER_VERIFICATION_ENABLED:
            verification_flag_field = get_user_setting('VERIFICATION_FLAG_FIELD')
            kwargs[verification_flag_field] = False
            email_field_name = get_user_email_field_name()
            if (email_field_name not in serializer.validated_data
                    or not serializer.validated_data[email_field_name]):
                raise UserWithoutEmailNonverifiable()

        with transaction.atomic():
            user = serializer.save(**kwargs)
            if registration_settings.REGISTER_VERIFICATION_ENABLED:
                email_sender = registration_settings.REGISTER_VERIFICATION_EMAIL_SENDER
                email_sender(request, user)

        signals.user_registered.send(sender=None, user=user, request=request)
        output_serializer_class = registration_settings.REGISTER_OUTPUT_SERIALIZER_CLASS
        output_serializer = output_serializer_class(
            instance=user,
            context={'request': request},
        )
        user_data = output_serializer.data
        return Response(user_data, status=status.HTTP_201_CREATED)

    def get_serializer_class(self) -> Type[Serializer]:
        return registration_settings.REGISTER_SERIALIZER_CLASS


register = RegisterView.as_view()


class VerifyRegistrationSerializer(serializers.Serializer):  # noqa: E501 pylint: disable=abstract-method
    user_id = serializers.CharField(required=True)
    timestamp = serializers.IntegerField(required=True)
    signature = serializers.CharField(required=True)


class VerifyRegistrationView(BaseAPIView):
    permission_classes = registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES

    def post(self, request: Request) -> Response:
        """
        Verify registration via signature.
        """
        user = process_verify_registration_data(
            request.data,
            serializer_context=self.get_serializer_context(),
        )
        signals.user_activated.send(sender=None, user=user, request=request)
        extra_data = None
        if registration_settings.REGISTER_VERIFICATION_AUTO_LOGIN:
            extra_data = perform_login(request, user)
        return get_ok_response(_("User verified successfully"), extra_data=extra_data)


verify_registration = VerifyRegistrationView.as_view()


def process_verify_registration_data(input_data, serializer_context=None):
    if serializer_context is None:
        serializer_context = {}
    if not registration_settings.REGISTER_VERIFICATION_ENABLED:
        raise Http404()
    serializer = VerifyRegistrationSerializer(
        data=input_data,
        context=serializer_context,
    )
    serializer.is_valid(raise_exception=True)

    data = serializer.validated_data
    # We use the signer only for verification, therefore we don't need a base_url and
    # may set strict=False
    signer = RegisterSigner(data, strict=False)
    verify_signer_or_bad_request(signer)

    verification_flag_field = get_user_setting('VERIFICATION_FLAG_FIELD')
    user = get_user_by_verification_id(data['user_id'], require_verified=False)
    setattr(user, verification_flag_field, True)
    user.save()

    return user



from typing import Any, Dict, Optional, Type

from django.http import Http404
from django.utils.translation import gettext as _
from rest_framework import permissions, serializers
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import Serializer

from rest_registration import signals
from rest_registration.api.views.base import BaseAPIView
from rest_registration.exceptions import EmailAlreadyRegistered
from rest_registration.settings import registration_settings
from rest_registration.signers.register_email import RegisterEmailSigner
from rest_registration.utils.responses import get_ok_response
from rest_registration.utils.users import (
    get_user_by_verification_id,
    get_user_email_field_name,
    is_user_email_field_unique,
    user_with_email_exists
)
from rest_registration.utils.verification import verify_signer_or_bad_request


class RegisterEmailView(BaseAPIView):
    permission_classes = [permissions.IsAuthenticated]

    def post(self, request: Request) -> Response:
        '''
        Register new email.
        '''
        user = request.user
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        email = serializer.validated_data['email']
        email_already_used = (
            is_user_email_field_unique()
            and user_with_email_exists(email))

        if registration_settings.REGISTER_EMAIL_VERIFICATION_ENABLED:
            email_sender = registration_settings.REGISTER_EMAIL_VERIFICATION_EMAIL_SENDER  # noqa: E501
            email_sender(request, user, email, email_already_used=email_already_used)
        else:
            if email_already_used:
                raise EmailAlreadyRegistered()

            email_field_name = get_user_email_field_name()
            old_email = getattr(user, email_field_name)
            setattr(user, email_field_name, email)
            user.save()
            signals.user_changed_email.send(
                sender=None,
                user=user,
                new_email=email,
                old_email=old_email,
                request=request,
            )

        return get_ok_response(_("Register email link email sent"))

    def get_serializer_class(self) -> Type[Serializer]:
        return registration_settings.REGISTER_EMAIL_SERIALIZER_CLASS


register_email = RegisterEmailView.as_view()


class VerifyEmailSerializer(serializers.Serializer):  # noqa: E501 pylint: disable=abstract-method
    user_id = serializers.CharField(required=True)
    email = serializers.EmailField(required=True)
    timestamp = serializers.IntegerField(required=True)
    signature = serializers.CharField(required=True)


class VerifyEmailView(BaseAPIView):
    serializer_class = VerifyEmailSerializer
    permission_classes = registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES

    def post(self, request: Request) -> Response:
        '''
        Verify email via signature.
        '''
        process_verify_email_data(
            request.data,
            serializer_context=self.get_serializer_context(),
        )
        return get_ok_response(_("Email verified successfully"))


verify_email = VerifyEmailView.as_view()


def process_verify_email_data(
        input_data: Dict[str, Any],
        serializer_context: Optional[Dict[str, Any]] = None) -> None:
    if serializer_context is None:
        serializer_context = {}
    if not registration_settings.REGISTER_EMAIL_VERIFICATION_ENABLED:
        raise Http404()
    serializer = VerifyEmailSerializer(data=input_data, context=serializer_context)
    serializer.is_valid(raise_exception=True)

    data = serializer.validated_data
    # We use the signer only for verification, therefore we don't need a base_url and
    # may set strict=False
    signer = RegisterEmailSigner(data, strict=False)
    verify_signer_or_bad_request(signer)
    request = serializer_context.get('request')
    new_email = data['email']

    if is_user_email_field_unique() and user_with_email_exists(new_email):
        raise EmailAlreadyRegistered()

    email_field_name = get_user_email_field_name()
    user = get_user_by_verification_id(data['user_id'])
    old_email = getattr(user, email_field_name)
    setattr(user, email_field_name, new_email)
    user.save()

    signals.user_changed_email.send(
        sender=None,
        user=user,
        new_email=new_email,
        old_email=old_email,
        request=request,
    )



from typing import TYPE_CHECKING, Optional

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from rest_registration.settings import registration_settings

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser


def get_login_authentication_backend(
    user: Optional['AbstractBaseUser'] = None
) -> Optional[str]:
    if user and hasattr(user, 'backend'):
        return user.backend  # type: ignore
    backends = settings.AUTHENTICATION_BACKENDS
    default_login_backend = registration_settings.LOGIN_DEFAULT_SESSION_AUTHENTICATION_BACKEND  # noqa: E501
    if not backends:
        raise ImproperlyConfigured("No AUTHENTICATION_BACKENDS specified")
    if len(backends) == 1:
        return backends[0]
    if default_login_backend is not None and default_login_backend not in backends:
        raise ImproperlyConfigured(
            "LOGIN_DEFAULT_SESSION_AUTHENTICATION_BACKEND"
            " is not in AUTHENTICATION_BACKENDS")
    return default_login_backend



"""
Custom signals sent during the registration and activation processes.
"""

from django.dispatch import Signal

# A new user has registered.
user_registered = Signal()

# A user has activated his or her account.
user_activated = Signal()

# A user has verified his or her new email.
user_changed_email = Signal()



from typing import Type

from rest_framework import permissions
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import Serializer

from rest_registration.api.views.base import BaseAPIView
from rest_registration.settings import registration_settings


class ProfileView(BaseAPIView):
    permission_classes = [permissions.IsAuthenticated]

    def get_serializer_class(self) -> Type[Serializer]:
        return registration_settings.PROFILE_SERIALIZER_CLASS

    def get(self, request: Request) -> Response:
        serializer = self.get_serializer(instance=request.user)
        return Response(serializer.data)

    def post(self, request: Request) -> Response:
        return self._update_profile(request)

    def put(self, request: Request) -> Response:
        return self._update_profile(request)

    def patch(self, request: Request) -> Response:
        return self._update_profile(request, partial=True)

    def _update_profile(self, request: Request, partial: bool = False) -> Response:
        serializer = self.get_serializer(
            instance=request.user,
            data=request.data,
            partial=partial,
        )
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(serializer.data)


profile = ProfileView.as_view()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from django_stubs_ext import StrOrPromise

    # mypy uses typing_extensions by default (Py 3.8); when importing from typing
    # one will get the following error message in mypy:
    #
    #   Incompatible import of "Literal"
    #   (imported name has type "typing_extensions._SpecialForm", local name has type
    #   "typing._SpecialForm")
    from typing_extensions import Literal
else:
    try:
        from typing import Literal
    except ImportError:
        from typing_extensions import Literal  # noqa: F401
    StrOrPromise = str
